from os.path import exists
from os import remove
from rbtconfig.rbtexceptions import RbtConfigError
from re import match
import tarfile


class RbtCommonConfigParser:
    __EXPR = r'^(?!#)(\S+)(\s+)?=(\s+)?(\S+)'
    __CONFIG_EXPR = r'^[ |\t]*(\w?\S?\W?)*\n?$'

    def __init__(self, service_config_dir):
        self.__backup_path = '%s/backup.tar.gz' % service_config_dir

    def __get_lines(self, config_file):
        """
        Получить список строк из конфигурационного файла
        :return: list
        """
        if not exists(config_file):
            raise RbtConfigError('Конфигурационный файл \'%s\' не найден' % config_file)
        return open(config_file).readlines()

    @staticmethod
    def __is_match(regexp, line, target):
        """
        Проверяет, совпадают ли строка из файла с :param target
        :param regexp: regexp
        :param line: str
        :param target: str
        :return: bool
        """
        try:
            result = match(regexp, line).group().replace('\t', '').replace('\n', '')
            return result == target
        except AttributeError:
            return False

    @staticmethod
    def __is_match_with_target_list(regexp, line, target_list):
        """
        Проверяет, совпадает ли строка из файла хотя бы с одной из :param target_list
        :param regexp: regexp
        :param line: str
        :param target_list: list[str]
        :return: bool
        """
        try:
            result = match(regexp, line).group().replace('\t', '').replace('\n', '')
            return result in target_list
        except AttributeError:
            return False

    def append_file_to_backup(self, path):
        """
        Добавляет файл в бекап
        :param path: str
        :return: None
        """
        try:
            tar = tarfile.open(self.__backup_path, mode='a')
        except tarfile.ReadError:
            tar = tarfile.open(self.__backup_path, mode='w')
        if path.lstrip('/') not in tar.getnames():
            tar.add(path)
        tar.close()

    def restore_backup(self):
        """
        Восстанавливает backup
        :return: None
        """
        if exists(self.__backup_path):
            tar = tarfile.open(self.__backup_path, mode='r')
            for file in tar.getnames():
                remove('/%s' % file)
                tar.extract(file, '/')
            remove(self.__backup_path)

    def set_or_replace_key_in_config(self, config_file, key, value, need_escape_strings=False, remove_spaces=False):
        """
        Устанавливает или заменяет ключ :param key значением :param value в конфигурационном файле :param config_file
        Если параметр не найден, добавляет его в конец конфигурационного файла :param config_file
        :param config_file: str
        :param key: str
        :param value: str
        :param need_escape_strings: нужно ли заключать в кавычки :param value: bool
        :param remove_spaces: нужно ли удалять пробелы между key = value: bool
        """
        lines = self.__get_lines(config_file)
        self.append_file_to_backup(config_file)
        results = []
        found = False
        for line in lines:
            groups = match(self.__EXPR, line)
            if groups:
                groups = list(groups.groups())
                groups = [element for element in groups if element not in [' ', '', None]]
                found_key = groups[0]
                found_value = groups[1]
                if found_key == key:
                    found = True
                    comments = str()
                    if '#' in line:
                        comments = line[line.index('#'):]
                    if not found_value == value:
                        if need_escape_strings:
                            value = '\'%s\'' % value
                        if remove_spaces:
                            if comments:
                                line = '%s=%s %s\n' % (found_key, value, comments)
                            else:
                                line = '%s=%s\n' % (found_key, value)
                        else:
                            if comments:
                                line = '%s = %s %s\n' % (found_key, value, comments)
                            else:
                                line = '%s = %s\n' % (found_key, value)
                    else:
                        if need_escape_strings:
                            value = '\'%s\'' % value
                        if remove_spaces:
                            line = '%s=%s\n' % (found_key, value)
                        else:
                            line = '%s = %s\n' % (found_key, value)
            results.append(line)
        if not found:
            if need_escape_strings:
                value = '\'%s\'' % value
            if remove_spaces:
                results.append('%s=%s\n' % (key, value))
            else:
                results.append('%s = %s\n' % (key, value))
        open(config_file, mode='w').writelines(results)

    def config_file_contains_key_value(self, config_file, key, value, need_escape_string=False):
        """
        Конфигурационный файл :param config_file содержит :param key равного :param value?
        :param config_file: str
        :param key: str
        :param value: str
        :param need_escape_string: нужно ли заключать в кавычки :param value: bool
        :return: bool
        """
        lines = self.__get_lines(config_file)
        if need_escape_string:
            value = '\'%s\'' % value
        for line in lines:
            groups = match(self.__EXPR, line)
            if groups:
                groups = list(groups.groups())
                groups = [element for element in groups if element not in [' ', '', None]]
                found_key = groups[0]
                found_value = groups[1]
                if found_key == key:
                    if not found_value == value:
                        return False
                    else:
                        return True
        return False

    def config_file_contains_only_one_row(self, config_file, row):
        """
        Возвращает True, если конфигурационный файл :param config_file содержит только одну строку :param row
        :param config_file: str
        :param row: str
        :return: bool
        """
        lines = self.__get_lines(config_file)
        if not len(lines) == 1:
            return False
        return row == lines[0].rstrip('\n')

    def insert_row_to_config_file(self, config_file, row, mode='w'):
        if not exists(config_file):
            raise RbtConfigError('Конфигурационный файл \'%s\' не найден' % config_file)
        open(config_file, mode).writelines([row])

    def add_after(self, config_file, rows, after=None):
        """
        Добавляет строки :param rows после строки :param after в конфигурационный файл :param config_file
        :param config_file: str
        :param after: str если None -- добавляется в конец файла
        :param rows: list[str]
        """
        lines = self.__get_lines(config_file)
        self.append_file_to_backup(config_file)
        results = []
        if after is not None:
            found = False
            for line in lines:
                results.append(line)
                if RbtCommonConfigParser.__is_match(self.__CONFIG_EXPR, line, after):
                    found = True
                    for row in rows:
                        if row in lines:
                            continue
                        if not row.endswith('\n'):
                            row += '\n'
                        results += row
            if not found:
                raise RbtConfigError('Строка \'%s\' не найдена в конфигурационном файле %s' % (after, config_file))
        else:
            results += lines
            for row in rows:
                if not row.endswith('\n'):
                    row += '\n'
                results += row
        open(config_file, mode='w').writelines(results)

    def replace_string(self, config_file, source, destination):
        """
        Заменяет строку :param source на :param destination в конфигурационном файле :param config_file
        :param config_file: str
        :param source: str
        :param destination: str
        """
        lines = self.__get_lines(config_file)
        self.append_file_to_backup(config_file)
        results = []
        for line in lines:
            if line.rstrip('\n') == source:
                results.append(destination)
            else:
                results.append(line)
        open(config_file, mode='w').writelines(results)

    def config_file_contains_rows(self, config_file, rows):
        """
        Конфигурационный файл :param config_file содержит строки :param rows?
        :param config_file: имя конфигурационного файла (srt)
        :param rows: list[str]
        :return: bool
        """
        lines = self.__get_lines(config_file)
        lines = [line.replace('\n', '').replace('\t', '') for line in lines]
        for row in rows:
            if row not in lines:
                return False
        return True

    def remove_rows_from_config(self, config_file, rows):
        """
        Удаляет строки :param rows из :param config_file
        :param config_file: имя конфигурационного файла (str)
        :param rows: list[str]
        """
        lines = self.__get_lines(config_file)
        for line in lines:
            if RbtCommonConfigParser.__is_match_with_target_list(self.__CONFIG_EXPR, line, rows):
                lines.remove(line)
        open(config_file, mode='w').writelines(lines)

    def write_file_with_content(self, config_file, rows):
        """
        Создает файл по пути :param config_file с содержимым :param rows
        :param config_file: str
        :param rows: list[str]
        :return: None
        """
        rows = [row + '\n' for row in rows if not row.endswith('\n')]
        open(config_file, mode='w').writelines(rows)
